
/************************************************************************/
/*                                                                      */
/*      linda3.h : include du systeme de communication MA               */
/*      INCLUS EGALEMENT LES PROTO DU SERVEUR DE DEBUG                  */
/*                                                                      */
/*      Copyright (C) 1991/92/93/94/95  LGIL puis LAMIH-GIL             */
/*                                                                      */
/************************************************************************/
/* Auteurs:
	V1.xx	Arnaud Hervas
		Premier design, Acces mono-associatif realise avec
		tres peu d'informations sur Helios. Peu fiable
		pour cette raison.
	V2.xx	Alain Gibaud
		Deuxieme design, tenant compte des info figurant 
		dans le livre "The Helios parallel O.S, 2nd version"
		-reecriture de 50% du code
		-introduction du mecanisme de trace
		-mecanisme d'horloge globale sommaire
		-ecriture d'un squelette de debugger
		-librairie d'acces supportant les clients
		en multi-thread 
	V3.xx	Alain Gibaud
		-developpement et implementation de la notion de typage
		-acces multi-associatif
		-implementation de eval()
		-documentation
		
  On egalement collabore a cette implementation:
  
  	Philippe Thomin: 
  		-utilisation de Linda et aide a la comprehension 
  		et a la gestion d'Helios. 
  		Maintenance (c.a.d. bricolages) hardware.
  	Regis Carlier:
  		-debugger graphique sous X-Window
  		pour la version 2.xx. (Non maintenu en V3.xx) 
  	Isabelle Gamain:
  		-debugger non-graphique pour la version 2.xx
  		(non maintenu en V3.xx car les principes de base de 
  		ce debugger sont rendus obsoletes par la notion de
  		tuple type)
  		-depouillement des traces sous la forme d'une
  		sortie PostScript. (Maintenable en V3.xx.)
*/
/* Deficiences connues:
	1) Le code de retour MEMORY_ERR est gere de maniere inconsistante 
	dans la librairie d'acces et le serveur. D'une maniere generale
	le bon comportement en cas de manque de memoire necessite une 
	modification attentive des sources.
	
	2) L'arret d'un serveur de processus laisse sur le serveur
	T-Linda des processus orphelins. Il est donc pour l'instant
	necessaire d'arreter le serveur Linda apres avoir arrete un serveur
	de processus.
*/
#ifndef __Linda_h
#define __Linda_h

#include <syslib.h>
#include <gsp.h>
#include <root.h>
#include <servlib.h>
#include <sem.h>
#include <nonansi.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>
#include <codes.h>
#include <stdlib.h>
#include <stdio.h>

#define LINDA_ID        "Helios Linda Server - Version 3.5 - (C) LGIL 91-95\n"
#define LINDA_VERSION   3.5


/* Nom du serveur et parametres de hashing
*******************************************/

#define LSName          "linda"
#define MaxLindaServer  16
#define _hpage          16
#define _hsize          4093

/* defines speciaux
********************/

#define MAX_TPISIZE     16	/* Taille maxi d'un identificateur      */
#define MAX_PSIZE       65500	/* Taille maxi d'un paquet de donnees   */

/* Codes des requetes
**********************/

#define Msg_LndPutWord          (FG_PrivateFirst | 0x00)
#define Msg_LndPut              (FG_PrivateFirst | 0x01)
#define Msg_LndGet              (FG_PrivateFirst | 0x02)
#define Msg_LndRead             (FG_PrivateFirst | 0x03)
#define Msg_LndGetFast          (FG_PrivateFirst | 0x04)
#define Msg_LndReadFast         (FG_PrivateFirst | 0x05)
#define Msg_LndGetU             (FG_PrivateFirst | 0x06)
#define Msg_LndReadU            (FG_PrivateFirst | 0x07)
#define Msg_LndGetFastU         (FG_PrivateFirst | 0x08)
#define Msg_LndReadFastU        (FG_PrivateFirst | 0x09)
#define Msg_KillServer          (FG_PrivateFirst | 0x0A)
#define Msg_LndDbgLevel		(FG_PrivateFirst | 0x0B)
#define Msg_LndDbgNop		(FG_PrivateFirst | 0x0C)
#define Msg_LndSynchronize	(FG_PrivateFirst | 0x0D)
#define Msg_LndPrintTuples	(FG_PrivateFirst | 0x0E)
#define Msg_LndPutType		(FG_PrivateFirst | 0x0F)
#define Msg_LndGetTypeFast	(FG_PrivateFirst | 0x10)
#define Msg_LndOut		(FG_PrivateFirst | 0x11)
#define Msg_LndIn		(FG_PrivateFirst | 0x12)
#define Msg_LndRd		(FG_PrivateFirst | 0x13)

/* Type definition
*******************/
/* ------------------------------------------------------------------------- */

typedef byte TupleId[MAX_TPISIZE];	 /* nom d'un tuple */

/* ------------------------------------------------------------------------- */
typedef struct TupleMsg			/* Bloc de donnees      */
{										  /* d'une requete        */
	byte      id[MAX_TPISIZE];	/* Identificateur       */
	word      size;			/* Taille ou valeur si  requete putword */
	long	  selec_size ;		  
	/* taille du bloc de selection (pour requete in() et rd()) 
	** ou bien flag 'duplicate' pour creation de type 
	** 1 == duplication de type autorise, 0 == interdit 
	*/
	
	/* **** Les informations suivantes sont utilisees par le debugger ****  */
	long      client_id;		/* client ayant emis la requete */
	long      emit_date;		/* heure (globale) d'emission */
	long      recep_date;		/* heure (globale) de reception */

}
TupleMsg , *tpmPtr;

/*
**	Definitions liees a la gestion des types et a la recherche
**	associative.
********************************************************************/

/* ------------------------------------------------------------------------- */
/*	Selecteur permettant la localisation des cles  de selection
	dans un tuple.
*/
	
typedef struct selector
{
	unsigned long nbelem ;	/* type et taille (en nb d'elem) codage 3+29 */
	long size ;		/* taille du champ (en byte) */
	long offset ;		/* decalage/ debut tuple */
	long arg_offset ;	/* decalage/ debut bloc des arguments reels */
	
} SELECTOR , *TupleSelector ;
 /*	L'entree No 0 d'une table de selecteurs	est speciale */
typedef struct selector0
{
	unsigned long nbelem ;
	long size ;	/* nb de selecteurs, SANS compter le selecteur 0 */
	struct 
	tuple_type *type_id ;	/* type local ou global correspondant
				** a cette table 
				*/
	long magic ;	/* nombre magique devant valoir MAGIC_SELECTOR */
} SELECTOR0 ;

/* ------------------------------------------------------------------------- */
/*	Description d'un noeud de l'arbre definissant un type utilisateur */
typedef struct decl
{
	short     type;		/* type de la declaration (voir #define(s) plus loin */
	long      taille;	/* taille en bytes */
	long      nbelem;	/* nb d'element, si tableau */
	long      offset;	/* offset / debut du tuple */
	struct decl
		 *frere,	/* declaration suivante, au meme niveau */
	         *fils;		/* description d'un bloc */
} DECL;

/* note: l'utilisation de calloc() garantit l'initialisation a 0 */
#define newDECL(t)	((DECL *)calloc((t), sizeof(DECL)))



/* ------------------------------------------------------------------------- */
typedef struct tuple_type
{
	char      name[MAX_TPISIZE]; /* nom du type */
	char     *fmt;			/* descriptif du type */
	struct decl *decl;		/* arbre de description du type
					** (obtenu par analyse de fmt)
					*/
	struct tuple_type *glob_handle;	/* handle du type global correspondant
					** au type local, ou NULL si il s'agit
					** d'un type global
					*/
	long      numdef;	        /* nombre de definition de ce type */
	struct tuple_type *next;	/* prochain dans la liste */
} TUPLE_TYPE , *TupleType ;

/* valeurs renvoyees par PutGlobalType() (et par le serveur) */
/* Pas de connexion au serveur */
#define NO_CONNECT      ((TupleType)-1)
/* PB de memoire  */
#define NO_MEMORY 	((TupleType)-2)
/* Le type defini n'a pas la meme definition que l'homonyme deja defini */
#define TYPE_MISMATCH	((TupleType)-3)
/* Le descripteur de format n'est pas syntaxiquement correct */
#define TYPE_BADDESC    ((TupleType)-4)

/* ----------------------------------------------------------------------- 
	structure passee au gardien d'evaluation eval_gard() 
	par register_process() 
*/

typedef struct eval_process
{
	char name[MAX_TPISIZE] ;	/* nom process */
	unsigned long stack_size ;	/* taille pile pour ce process */
	VoidFnPtr pf ;			/* adr fonction */
	int flag ;			/* modificateur pour ce process */
} EVAL_PROCESS ;

/* valeurs permises pour flag */
/* une seule evaluation a la fois par serveur (performance) */
#define EVL_UNIQUE 0
/* evaluation multiples simultanees (concurrence) */
#define EVL_MULTI  1

/* valeurs renvoyees par lput, lget, lread (+ versions u et nb) */
#define CONNECT_ERR (-1) /* code fiable */
#define MEMORY_ERR   0	/* non fiable - necessite de revoir le code */
#define NO_TUPLE_ERR 0	 /* code fiable */


/* ------------------------------------------------------------------------- */
typedef struct TupleHeader		/* Entete d'un tuple    */
{				  	/* Utilises dans les    */
	byte      id[MAX_TPISIZE];	/* listes de la table   */
	word      size;			/* de hachage           */
	byte     *data;
	TUPLE_TYPE *type_handle;	/* NULL si non type, adr. du type autrement */
} TupleHeader, *tphPtr;



/* ------------------------------------------------------------------------- */
typedef struct TupleList		/* Liste d'entetes      */
{
	word      list_size;		/* Nbre total d'entetes */
	word      list_alloc;		/* Nbre d'ent. utilises */
	tphPtr    list;			/* Tableau d'entetes    */
	Semaphore lock;			/* Sem. d'acces         */
	Semaphore waiting;		/* Sem. lecture bloquee */
	word      numwait;		/* Nbre de taches       
					** bloquees en lecture  */
	TUPLE_TYPE *type_table;		/* table des types pour cette entree */
} TupleList, *tplPtr;



/* ------------------------------------------------------------------------- */
typedef struct TupleSpace		/* Table de hachage     */
{
	tplPtr    hash_table;		/* 1 point d'entree     */
	Semaphore lock;			/* = 1 liste d'entetes  */

} TupleSpace, *tpsPtr;



/* ------------------------------------------------------------------------- */
/* 	map des serveurs deja presents, renvoyee par la fonction
	de connection: il s'agit d'une COPIE des donnees internes
	de la librairie client
*/

typedef struct LndServMap
{
	Port      lndport[MaxLindaServer];	/* map des port ouverts (NullPort si ferme) */
	Port      lndactv[MaxLindaServer];	/* meme table sous forme compacte */
	int       numserver;			/* nombre de serveurs trouves */
}

LNDSERVMAP;



/* ------------------------------------------------------------------------- */

/* Fonctions privees du gestionnaire de memoire (Serveur)
*********************************************************/

extern word InitTupleSpace(tpsPtr);
extern word GetHashCode(byte *);
extern bool MatchID(byte *, byte *);
extern tphPtr GetNewSlot(tplPtr);
extern void MsgHandler(MsgBuf *);

extern void LindaPutWord(MCB *);
extern void LindaPut(MCB *);
extern void LindaGet(MCB *);
extern void LindaRead(MCB *);
extern void LindaGetU(MCB *);
extern void LindaReadU(MCB *);
extern void LindaGetRead(MCB *, bool, bool);
extern void LindaGetFast(MCB *);
extern void LindaReadFast(MCB *);
extern void LindaGetFastU(MCB *);
extern void LindaReadFastU(MCB *);
extern void LindaGetReadFast(MCB *, bool, bool);
extern void LindaKill(MCB *);
extern void LindaChangeDbgLevel(MCB *);
extern void LindaNop(MCB *);
extern void LindaSynchronize(MCB *);
extern void LindaPrint(MCB *);
extern void LindaPutType(MCB *);
extern void LindaGetTypeFast(MCB *);
extern void LindaOut(MCB *);
extern void LindaIn(MCB *);
extern void LindaRd(MCB *);

/* Fonctions privees de la librairie d'acces au serv std
********************************************************/

extern int LindaConnect(LNDSERVMAP * map);
extern void LindaDeconnect(void);
extern Port SelectServer(byte *);
extern void MakeTupleID(byte *, byte *);
extern int LindaDbg(int mode, int num);

/* un code special pour 'num' : */
#define ALLSERVERS	-2
extern void LindaSynchro(void);
extern long LindaGlobalClock(void);
extern int LindaConnected(void);
extern int LindaClientNum(void);
extern char *LindaRequestName(long);
extern int LindaPrintTuples(int);
extern TUPLE_TYPE *PutGlobalType(char *name, char *fmt, int duplicate);
extern TUPLE_TYPE **FindLocalTypeByName(char *name, TUPLE_TYPE ** root);
extern TUPLE_TYPE **FindLocalType(TUPLE_TYPE *t, TUPLE_TYPE **root) ;
extern TUPLE_TYPE *DeleteLocalType(TUPLE_TYPE *t , TUPLE_TYPE ** root);
extern int analyse_fmt(char *fmt, DECL ** d);
extern void MakeMCB( MCB* mcb, 
		word fnrc, 
		Port dest,
		Port reply,
		byte *data, unsigned short datasize,
		word *control, unsigned char contsize,
		word timeout,
		unsigned char flags ) ;
extern void 	PrintSelector( SELECTOR *s) ;
extern TUPLE_TYPE *NewTupleType(void) ;
extern TUPLE_TYPE *def_tuple_type(char *name, char *fmt, int duplicate, va_list) ;
extern TUPLE_TYPE *global_id(TUPLE_TYPE *id) ;
extern void eval_guard( EVAL_PROCESS *pf ) ;

/* Fonctions Publiques d'acces (SERVEUR STANDARD)
*******************************/
/* 	les fonctions suivantes
	(ainsi que leur homologues dediees au serveur debug)
	renvoient les nombre d'octets lus (ou ecrits)
	ou un des deux codes d'erreur negatifs suivants:
*/
#define NOSERVER	(-1)
#define NOSPACE		(-2)	/* non fiable (de meme que MEMORY_ERR) */

/*	Fonctions de niveau 1
	---------------------
*/

extern int lputw(byte *, int);
extern int lput(byte *, void *, int);
extern int lget(byte *, void *);
extern int lread(byte *, void *);
extern int lgetnb(byte *, void *);
extern int lreadnb(byte *, void *);
extern int lgetu(byte *, void **);
extern int lreadu(byte *, void **);
extern int lgetunb(byte *, void **);
extern int lreadunb(byte *, void **);

/*	Fonctions de niveau 2 (tuple types et acces associatif)
	-------------------------------------------------------
*/
extern TUPLE_TYPE *DefineTupleType(char *name, char *fmt, ...) ;
extern TUPLE_TYPE *DefineGlobalTupleType(char *name, char *fmt, ...) ;

extern int 	 UndefTupleType(TUPLE_TYPE *t);
extern int out(TUPLE_TYPE * type, void *data);
extern TupleSelector CreateTupleSelector( TupleType type_id, char *mask, ... ) ; 
extern int in( SELECTOR *selec, void *data ) ;
extern int rd( SELECTOR *selec, void *data ) ;
/* In et Rd sont equivalents a in et rd pour l'instant */
#define In(a,b) in(a,b)
#define Rd(a,b) rd(a,b)
extern char *strtoarray( char *to, char *from, size_t size) ;
#define MakeByteArray(a,b) strtoarray((a),(b),sizeof(a) ) 
extern void FillTuple( void *tuple, TupleSelector s, ...) ;
extern int Eval( char *procname, TUPLE_TYPE **tt, int ntt, void *data, long datasiz) ;
extern int RegisterProcess( long stack_size , char *name , VoidFnPtr fp , int flag ) ;
extern int EndEval( TUPLE_TYPE **t) ;


/* Fonctions privees de la librairie d'acces au serv DEBUG
********************************************************/

extern int LindaConnectD(LNDSERVMAP * map);
extern void LindaDeconnectD(void);
extern Port SelectServerD(byte *);
extern int LindaDbgD(int, int);
extern int LindaConnectedD(void);
extern int LindaPrintTuplesD(int);
extern TUPLE_TYPE *PutGlobalTypeD(char *name, char *fmt, int duplicate);



/* Fonctions Publiques d'acces (SERVEUR DEBUG)
*******************************/

extern int lputwD(byte *, int);
extern int lputD(byte *, void *, int);
extern int lgetD(byte *, void *);
extern int lreadD(byte *, void *);
extern int lgetnbD(byte *, void *);
extern int lreadnbD(byte *, void *);
extern int lgetuD(byte *, void **);
extern int lreaduD(byte *, void **);
extern int lgetunbD(byte *, void **);
extern int lreadunbD(byte *, void **);

/*	Fonctions de niveau 2 
	---------------------
*/
extern TUPLE_TYPE *DefineTupleTypeD(char *name, char *fmt, ...) ;
extern TUPLE_TYPE *DefineGlobalTupleTypeD(char *name, char *fmt, ...) ;
extern int 	 UndefTupleTypeD(TUPLE_TYPE *tt);
extern TupleSelector CreateTupleSelectorD( TupleType type_id, char *mask ) ; 
extern int outD(TupleType type, void *data);
extern int inD( TupleSelector selec, void *data ) ;
extern int rdD( TupleSelector selec, void *data ) ;
/* In et Rd sont equivalents a in et rd pour l'instant */
#define InD(a,b) inD(a,b)
#define RdD(a,b) rdD(a,b)
/* Note: Eval() et les fct qui lui sont liees n'existent 
** dans la librairie DEBUG
*/
/* -------------------------------------------------------------
**	Definit la structure du tuple informant le debugger qu'une requete
**	a ete recue par un serveur
*/
typedef struct DbgRequest
{
	long      code_requete;		  /* code du type Msg_Lnd... (voir plus haut) */
	long      client_id;			  /* id logique du client */
	long      server_id;			  /* numero du serveur concerne */
	long      recep_date;		  /* date de reception de la requete par le serveur */
	long      emit_date;
	char      tuple_name[MAX_TPISIZE];	 /* nom du tuple */
	char      tuple_ack[MAX_TPISIZE]; /* nom tuple d'acquittement */
	/* NOTE:    Le contenu du tuple n'est PAS encore Gere */
}

DBGREQUEST;

#endif
/* -----------------------------------------------------------------------
	Nombres magiques destines a authentifier 
	les structures allouees dynamiquement
*/
/* identifie un selecteur */
#define MAGIC_SELECTOR		0xCE231AF8L
/* identifie une table de types, passee a un process EVALue */
#define MAGIC_TYPE_TABLE	0xFE45012DL
	
/* ------------------------------------------------------ 
	constantes utiles pour initialiser les MCB 
*/
#define NO_PRESERVE	0
#define PRESERVE 	MsgHdr_Flags_preserve
#define INFINI 		-1

/* ------------------------------------------------------
	constantes renvoyees lors de la destruction d'un type
*/
#define NO_SUCH_TYPE	-1
#define HERITED_TYPE	-2

/* -------------------------------------------------------
	Ce semaphore et ses macro associees
	permettent de verouiller l'ecran, en mode
	multicache (usage debug)
*/

extern Semaphore _screenlock ;

#define SCREEN   Wait(&_screenlock), fflush(stdout) 
#define NOSCREEN fflush(stdout),Signal(&_screenlock) 

